import tkinter as tk
from tkinter import *
from tkinter import ttk
from idlelib.tooltip import Hovertip
import threading

from eo_man import LOGGER

from ..controller.app_bus import AppBus, AppBusEventType
from ..controller.serial_controller import SerialController
from ..data.data_manager import DataManager
from ..data import data_helper

from eltakobus.message import *
from eltakobus.eep import *
from eltakobus.util import *

from ..data.const import get_display_names, GATEWAY_DISPLAY_NAMES, GatewayDeviceType

class SerialConnectionBar():

    def __init__(self, main: Tk, app_bus:AppBus, data_manager:DataManager, serial_controller:SerialController, row:int):
        self.main = main
        self.app_bus = app_bus
        self.data_manager = data_manager
        self.serial_cntr = serial_controller

        f = LabelFrame(main, text="Serial Connection", bd=1)#, relief=SUNKEN)
        f.grid(row=row, column=0, columnspan=1, sticky=W+E+N+S, pady=(0,2), padx=2)

        self.b_detect = ttk.Button(f, text="Detect Serial Ports")
        self.b_detect.pack(side=tk.LEFT, padx=(5, 5), pady=5)
        self.b_detect.config(command=lambda : self.detect_serial_ports_command(force_reload=True) )
        Hovertip(self.b_detect,"Serial port detection is sometime unstable. Please try again if device was not detected.",300)

        l = Label(f, text="Gateway Type: ")
        l.pack(side=tk.LEFT, padx=(0, 5), pady=5)

        self.cb_device_type = ttk.Combobox(f, state="readonly", width="18") 
        self.cb_device_type.pack(side=tk.LEFT, padx=(0, 5), pady=5)
        self.cb_device_type['values'] = get_display_names() # ['FAM14', 'FGW14-USB', 'FAM-USB', 'USB300', 'LAN Gateway']
        self.cb_device_type.set(self.cb_device_type['values'][0])
        self.cb_device_type.bind('<<ComboboxSelected>>', self.on_device_type_changed)

        l = ttk.Label(f, text="Serial Port: ")
        l.pack(side=tk.LEFT, padx=(0, 5), pady=5)

        self.cb_serial_ports = ttk.Combobox(f, state="readonly", width="14") 
        self.cb_serial_ports.pack(side=tk.LEFT, padx=(0, 5), pady=5)

        self.b_connect = ttk.Button(f, text="Connect", state=DISABLED, command=self.toggle_serial_connection_command)
        self.b_connect.pack(side=tk.LEFT, padx=(0, 5), pady=5)

        s = ttk.Separator(f, orient=VERTICAL )
        s.pack(side=tk.LEFT, padx=(0,5), pady=0, fill="y")

        self.b_scan = ttk.Button(f, text="Scan for devices", state=DISABLED, command=self.scan_for_devices)
        self.b_scan.pack(side=tk.LEFT, padx=(0, 5), pady=5)

        self.overwrite = tk.BooleanVar()
        self.cb_overwrite = ttk.Checkbutton(f, text="Overwrite existing values", variable=self.overwrite)
        self.cb_overwrite.pack(side=tk.LEFT, padx=(0, 5), pady=5)

        s = ttk.Separator(f, orient=VERTICAL )
        s.pack(side=tk.LEFT, padx=(0,5), pady=0, fill="y")

        text  = "Ensures sender configuration for Home Assistant is written into device memory.\n"
        text += "* Gateways will be added when being once connected.\n"
        text += "* Only devices connected to FAM14 via wire will be updated.\n"
        text += "* Button will be enabled when FAM14 is connected."

        l = ttk.Label(f, text="Program HA senders into devices: ")
        Hovertip(l, text, 300)
        l.pack(side=tk.LEFT, padx=(0, 5), pady=5)

        self.cb_gateways_for_HA = ttk.Combobox(f, state="readonly", width="24") 
        Hovertip(self.cb_gateways_for_HA, text, 300)
        self.cb_gateways_for_HA.pack(side=tk.LEFT, padx=(0, 5), pady=5)

        self.b_sync_ha_sender = ttk.Button(f, text="Write to devices", state=DISABLED, command=self.write_ha_senders_to_devices)
        Hovertip(self.b_sync_ha_sender, text, 300)
        self.b_sync_ha_sender.pack(side=tk.LEFT, padx=(0, 5), pady=5)

        # # if connected via fam14 force to get status update message
        # b = ttk.Button(f, text="Send Poll", command=lambda: self.serial_cntr.send_message(EltakoPollForced(5)))
        # b.pack(side=tk.LEFT, padx=(0, 5), pady=5)

        self.app_bus.add_event_handler(AppBusEventType.CONNECTION_STATUS_CHANGE, self.is_connected_handler)
        self.app_bus.add_event_handler(AppBusEventType.DEVICE_SCAN_STATUS, self.device_scan_status_handler)
        self.app_bus.add_event_handler(AppBusEventType.WRITE_SENDER_IDS_TO_DEVICES_STATUS, self.device_scan_status_handler)
        self.app_bus.add_event_handler(AppBusEventType.WINDOW_LOADED, self.on_window_loaded)
        self.app_bus.add_event_handler(AppBusEventType.UPDATE_DEVICE_REPRESENTATION, self.update_cb_gateways_for_HA)
        self.app_bus.add_event_handler(AppBusEventType.UPDATE_SENSOR_REPRESENTATION, self.update_cb_gateways_for_HA)
        

    def update_cb_gateways_for_HA(self, event=None):
        gateways = []
        for d in self.data_manager.devices.values():
            if d.is_gateway():
                gateways.append(f"{d.device_type.replace(' (Wireless Transceiver)', '')} ({d.external_id})")
        
        self.cb_gateways_for_HA['values'] = gateways
        if self.cb_gateways_for_HA.get() == '' and len(gateways) > 0:
            self.cb_gateways_for_HA.set(gateways[0])
        elif len(gateways) == 0:
            self.cb_gateways_for_HA.set('')

    def on_device_type_changed(self, event):
        self.cb_serial_ports['values'] = []
        self.cb_serial_ports.set('')

        self.detect_serial_ports_command()


    def write_ha_senders_to_devices(self):
        # get gateways and its base id
        baseId = self.cb_gateways_for_HA.get().split(' ')[-1].replace('(', '').replace(')', '')
        g = self.data_manager.devices.get(baseId, None)
        if g.is_wired_gateway():
            sender_base_id = '00-00-B0-00'
        else:
            sender_base_id = g.base_id

        # get sender list to be entered
        sender_list = {}
        for d in self.data_manager.devices.values():
            if d.base_id == self.serial_cntr.current_base_id and d.use_in_ha:
                if 'sender' in d.additional_fields:
                    sender_list[d.external_id] = dict(d.additional_fields)
                    sender_id = data_helper.a2i(sender_base_id) + data_helper.a2i(d.address)
                    sender_list[d.external_id]['sender']['id'] =  data_helper.a2s(sender_id)

        self.app_bus.fire_event(AppBusEventType.LOG_MESSAGE, {
            'msg': f'Write sender ids to devices for gateway {self.cb_gateways_for_HA.get()} with sender base id {sender_base_id} on bus with base id {self.serial_cntr.current_base_id}',
            'color': 'lightgreen'})
        self.serial_cntr.write_sender_id_to_devices(sender_list)
        

    def on_window_loaded(self, data):
        self.detect_serial_ports_command()


    def scan_for_devices(self):
        self.serial_cntr.scan_for_devices( self.overwrite.get() )


    def device_scan_status_handler(self, status:str):
        if status == 'FINISHED':
            self.is_connected_handler({'connected': self.serial_cntr.is_serial_connection_active()})
            self.main.config(cursor="")
            self.b_scan.config(state=NORMAL)
            self.b_connect.config(state=NORMAL)
            self.b_sync_ha_sender.config(state=NORMAL)
        if status == 'STARTED':
            self.b_scan.config(state=DISABLED)
            self.b_connect.config(state=DISABLED)
            self.main.config(cursor="watch")    #set cursor for waiting
            self.b_sync_ha_sender.config(state=DISABLED)


    def toggle_serial_connection_command(self):
        try:
            if not self.serial_cntr.is_serial_connection_active():
                self.b_detect.config(state=DISABLED)
                self.b_connect.config(state=DISABLED)
                self.b_scan.config(state=DISABLED)
                self.serial_cntr.establish_serial_connection(self.cb_serial_ports.get(), self.cb_device_type.get())
            else:
                self.serial_cntr.stop_serial_connection()
        except:
            # reset buttons
            self.is_connected_handler(data={'connected': self.serial_cntr.is_serial_connection_active()})
            LOGGER.exception("Was not able to detect serial ports.")


    def is_connected_handler(self, data:dict, skipp_serial_port_detection:bool=False):
        status = data.get('connected')
        
        self.main.config(cursor="")

        if status:
            self.b_connect.config(text="Disconnect", state=NORMAL)
            self.cb_serial_ports.config(state=DISABLED)
            self.b_detect.config(state=DISABLED)
            self.cb_device_type.config(state=DISABLED)
            
            if self.cb_device_type.get() == GATEWAY_DISPLAY_NAMES[GatewayDeviceType.EltakoFAM14] and self.serial_cntr.is_fam14_connection_active():
                self.b_scan.config(state=NORMAL)
                self.b_sync_ha_sender.config(state=NORMAL)
            else:
                self.b_scan.config(state=DISABLED)
                self.b_sync_ha_sender.config(state=DISABLED)

        else:
            self.b_connect.config(text="Connect", state=NORMAL)
            self.b_detect.config(state=NORMAL)
            self.cb_serial_ports.config(state="readonly")
            self.cb_device_type.config(state="readonly")
            self.b_scan.config(state=DISABLED)
            self.b_sync_ha_sender.config(state=DISABLED)
            if not skipp_serial_port_detection: self.detect_serial_ports_command()


    def detect_serial_ports_command(self, force_reload:bool=False):

        def detect_serial_ports():
            try:
                self.main.config(cursor="watch")    #set cursor for waiting
                self.b_detect.config(state=DISABLED)
                self.b_connect.config(state=DISABLED)
                self.cb_device_type.config(state=DISABLED)
                self.cb_serial_ports.config(state=DISABLED)
                serial_ports = self.serial_cntr.get_serial_ports(self.cb_device_type.get(), force_reload)
                self.b_detect.config(state=NORMAL)
                self.cb_device_type.config(state=NORMAL)
                self.cb_serial_ports['values'] = serial_ports
                if len(self.cb_serial_ports['values']) > 0:
                    self.cb_serial_ports.set(self.cb_serial_ports['values'][0])
                    self.b_connect.config(state=NORMAL)
                else:
                    self.b_connect.config(state=DISABLED)
                    self.cb_serial_ports.set('')
            except:
                # reset buttons
                LOGGER.exception("Was not able to detect serial ports.")
            else:
                self.is_connected_handler(data={'connected': False}, skipp_serial_port_detection=True)

        t = threading.Thread(target=detect_serial_ports)
        t.start()


from enum import Enum

class AppBusEventType(Enum):
    LOG_MESSAGE = 0                     # dict with keys: msg:str, color:str
    SERIAL_CALLBACK = 1                 # dict: msg:EltakoMessage base_id:str
    CONNECTION_STATUS_CHANGE = 2        # dict with keys: serial_port:str, baudrate:int, connected:bool
    DEVICE_ITERATION_PROGRESS = 3            # percentage 0..100 in float
    DEVICE_SCAN_STATUS = 4              # str: STARTED, FINISHED, DEVICE_DETECTED
    ASYNC_DEVICE_DETECTED = 5           # BusObject
    UPDATE_DEVICE_REPRESENTATION = 6    # busdevice
    UPDATE_SENSOR_REPRESENTATION = 7    # esp2  eltakomessage
    WINDOW_CLOSED = 8
    WINDOW_LOADED = 9
    SELECTED_DEVICE = 10                # device
    LOAD_FILE = 11
    WRITE_SENDER_IDS_TO_DEVICES_STATUS = 12
    SET_DATA_TABLE_FILTER = 13          # applies data filter to data table
    ADDED_DATA_TABLE_FILTER = 14        # adds data filter to application data
    REMOVED_DATA_TABLE_FILTER = 15      # remove data filter from application data
    ASYNC_TRANSCEIVER_DETECTED = 17      # type:str (FAM-USB), base_id:str 00-00-00-00
    SEND_MESSAGE_TEMPLATE_LIST_UPDATED = 18

class AppBus():

    def __init__(self) -> None:
        self.handler_count = 0

        for event_type in AppBusEventType:
            if event_type not in self._controller_event_handlers.keys():
                self._controller_event_handlers[event_type] = {}


    _controller_event_handlers={}
    def add_event_handler(self, event:AppBusEventType, handler) -> int:
        self.handler_count += 1
        self._controller_event_handlers[event][self.handler_count] = handler
        return self.handler_count
    
    def remove_event_handler_by_id(self, handler_id:int) -> None:
        for et in AppBusEventType:
            if handler_id in self._controller_event_handlers[et]:
                del self._controller_event_handlers[et][handler_id]
                break

    def fire_event(self, event:AppBusEventType, data) -> None:
        # print(f"[Controller] Fire event {event}")
        for h in self._controller_event_handlers[event].values(): h(data)

    async def async_fire_event(self, event:AppBusEventType, data) -> None:
        # print(f"[Controller] Fire async event {event}")
        for h in self._controller_event_handlers[event].values(): await h(data)


from ..controller.app_bus import AppBus, AppBusEventType
from . import data_helper 
from .device import Device 
from .application_data import ApplicationData
from .filter import DataFilter
from .const import *
from .app_info import ApplicationInfo as AppInfo
from .recorded_message import RecordedMessage
from .message_history import MessageHistoryEntry

from eltakobus.util import AddressExpression, b2s
from eltakobus.eep import EEP
from eltakobus.message import RPSMessage, Regular1BSMessage, Regular4BSMessage, EltakoMessage, EltakoWrappedRPS,EltakoWrapped4BS

class DataManager():
    """Manages EnOcean Devices"""

    def __init__(self, app_bus:AppBus):
        self.app_bus = app_bus
        self.app_bus.add_event_handler(AppBusEventType.SERIAL_CALLBACK, self._serial_callback_handler)
        self.app_bus.add_event_handler(AppBusEventType.ASYNC_DEVICE_DETECTED, self._async_device_detected_handler)
        self.app_bus.add_event_handler(AppBusEventType.LOAD_FILE, self._reset)
        self.app_bus.add_event_handler(AppBusEventType.SET_DATA_TABLE_FILTER, self.set_current_data_filter_handler)
        self.app_bus.add_event_handler(AppBusEventType.REMOVED_DATA_TABLE_FILTER, self.remove_current_data_filter_handler)
        self.app_bus.add_event_handler(AppBusEventType.ASYNC_TRANSCEIVER_DETECTED, self._async_transceiver_detected)
        self.app_bus.add_event_handler(AppBusEventType.SEND_MESSAGE_TEMPLATE_LIST_UPDATED, self.on_update_send_message_template_list)

        # devices
        self.devices:dict[str:Device] = {}

        # filter
        self.data_fitlers:dict[str:DataFilter] = {}
        self.selected_data_filter_name:DataFilter = None

        # recorded messages
        self.recoreded_messages:list[RecordedMessage] = []

        # message history
        self.send_message_template_list:list[MessageHistoryEntry] = None


    def set_current_data_filter_handler(self, filter:DataFilter):
        if filter is not None:
            self.selected_data_filter_name = filter.name
        else:
            self.selected_data_filter_name = None


    def remove_current_data_filter_handler(self, filter:DataFilter):
        if self.selected_data_filter_name is not None and self.selected_data_filter_name == filter.name:
            self.selected_data_filter_name = None


    def add_filter(self, filter:DataFilter) -> None:
        self.data_fitlers[filter.name] = filter

        self.app_bus.fire_event(AppBusEventType.ADDED_DATA_TABLE_FILTER, filter)


    def remove_filter(self, filter:DataFilter) -> None:
        if filter.name in self.data_fitlers.keys():
            del self.data_fitlers[filter.name]

        self.app_bus.fire_event(AppBusEventType.REMOVED_DATA_TABLE_FILTER, filter)


    def _reset(self, data):
        self.devices = {}


    def load_data_filters(self, filters:list[DataFilter]):
        for f in filters.values():
            self.add_filter(f)


    def load_devices(self, devices:dict[str:Device]):
        # load FAM14 first because of dependencies
        d_list =  [d.external_id for d in devices.values() if d.is_fam14()] 
        d_list += [d.external_id for d in devices.values() if not d.is_fam14()] 
        for key in d_list:
            device:Device = devices[key]
            self.devices[key] = device
            if device.is_bus_device():
                self.app_bus.fire_event(AppBusEventType.UPDATE_DEVICE_REPRESENTATION, device)
            else:
                self.app_bus.fire_event(AppBusEventType.UPDATE_SENSOR_REPRESENTATION, device)


    def load_application_data_from_file(self, filename:str):
        # if filename.endswith('.eodm'):
        #     app_data:ApplicationData = ApplicationData.read_from_file(filename)
        # elif filename.endswith('.yaml'):
        #     app_data:ApplicationData = ApplicationData.read_from_yaml_file(filename)
        app_data:ApplicationData = ApplicationData.read_from_yaml_file(filename)
        
        self.load_data_filters(app_data.data_filters)
        self.selected_data_filter_name = app_data.selected_data_filter_name
        if self.selected_data_filter_name is None or self.selected_data_filter_name == '': 
            self.app_bus.fire_event(AppBusEventType.SET_DATA_TABLE_FILTER, None)
        elif self.selected_data_filter_name not in self.data_fitlers.keys():
            self.app_bus.fire_event(AppBusEventType.SET_DATA_TABLE_FILTER, None)
        else:
            self.app_bus.fire_event(AppBusEventType.SET_DATA_TABLE_FILTER, self.data_fitlers[self.selected_data_filter_name])

        self.send_message_template_list = app_data.send_message_template_list

        self.recoreded_messages = app_data.recoreded_messages
        self.load_devices(app_data.devices)
        return app_data


    def write_application_data_to_file(self, filename:str):
        app_data = ApplicationData()
        app_data.application_version = AppInfo.get_version()
        app_data.data_filters = self.data_fitlers
        app_data.devices = self.devices
        app_data.selected_data_filter_name = self.selected_data_filter_name
        app_data.recoreded_messages = self.recoreded_messages
        app_data.send_message_template_list = self.send_message_template_list

        ApplicationData.write_to_yaml_file(filename, app_data)


    def _serial_callback_handler(self, data:dict):
        message:EltakoMessage = data['msg']
        current_base_id:str = data['base_id']
        gateway_id:str = data['gateway_id']

        if type(message) in [EltakoWrappedRPS,EltakoWrapped4BS, RPSMessage, Regular1BSMessage, Regular4BSMessage]:
            # for decentral devices
            if int.from_bytes(message.address, "big") > 0X0000FFFF:
                dev_address = b2s(message.address)
                # add message to list
                self.recoreded_messages.append(RecordedMessage(message, dev_address, gateway_id))
                # if device unknown add device to list
                if dev_address not in self.devices:
                    decentralized_device = Device.get_decentralized_device_by_telegram(message)
                    self.devices[dev_address] = decentralized_device
                    self.app_bus.fire_event(AppBusEventType.UPDATE_SENSOR_REPRESENTATION, decentralized_device)
            
            # for bus devices
            elif current_base_id:
                external_id = data_helper.a2s( int.from_bytes(AddressExpression.parse(current_base_id)[0], 'big') +  int.from_bytes(message.address, 'big') )
                # add message to list
                self.recoreded_messages.append(RecordedMessage(message, external_id, gateway_id))
                # if device unknown add device to list
                if external_id not in self.devices:
                    centralized_device = Device.get_centralized_device_by_telegram(message, current_base_id, external_id)
                    self.devices[centralized_device.external_id] = centralized_device
                    self.app_bus.fire_event(AppBusEventType.UPDATE_DEVICE_REPRESENTATION, centralized_device)


    async def _async_transceiver_detected(self, data):
        base_id = data['base_id']
        if data.get('type', None) is not None:
            gw_device = Device(address=base_id,
                             bus_device=False,
                             external_id=base_id,
                             base_id=base_id,
                             name=f"{data['type']} ({base_id})",
                             device_type=get_gateway_type_by_name(data['type']),
                             use_in_ha=True
                             )
            if data['type'] == GATEWAY_DISPLAY_NAMES[GatewayDeviceType.LAN]:
                gw_device.additional_fields['address'] = data['address']
            if gw_device.external_id not in self.devices:
                self.devices[gw_device.external_id] = gw_device
                self.app_bus.fire_event(AppBusEventType.UPDATE_SENSOR_REPRESENTATION, gw_device)



    async def _async_device_detected_handler(self, data):
        for channel in range(1, data['device'].size+1):
            bd:Device = await Device.async_get_bus_device_by_natvice_bus_object(data['device'], data['fam14'], channel)
            
            update = data['force_overwrite']
            update |= bd.external_id not in self.devices
            update |= bd.external_id in self.devices and self.devices[bd.external_id].device_type is None 
            update |= bd.external_id in self.devices and self.devices[bd.external_id].device_type == ''
            update |= bd.external_id in self.devices and self.devices[bd.external_id].device_type == 'unknown'

            if update:
                self.devices[bd.external_id] = bd
                self.app_bus.fire_event(AppBusEventType.UPDATE_DEVICE_REPRESENTATION, bd)

                for si in bd.memory_entries:
                    _bd:Device = await Device.async_get_decentralized_device_by_sensor_info(si, data['device'], data['fam14'], channel)

                    if _bd.external_id not in self.devices or not self.devices[_bd.external_id].bus_device:
                        self.devices[_bd.external_id] = _bd
                        self.app_bus.fire_event(AppBusEventType.UPDATE_SENSOR_REPRESENTATION, _bd)

                    # add device a second time with base id of FTD14
                    if bd.is_ftd14():                    
                        _bd:Device = Device.get_decentralized_device_by_sensor_info(si, bd.additional_fields['second base id'])
                    
                        if _bd.external_id not in self.devices or not self.devices[_bd.external_id].bus_device:
                            self.devices[_bd.external_id] = _bd
                            self.app_bus.fire_event(AppBusEventType.UPDATE_SENSOR_REPRESENTATION, _bd)
                
                # add features of device as own entity/device
                feature = Device.get_feature_as_device(bd)
                if feature is not None: 
                    self.devices[feature.external_id] = feature
                    self.app_bus.fire_event(AppBusEventType.UPDATE_DEVICE_REPRESENTATION, feature)

                # if a new gateway was detected check if there are already devices detected which should be moved as child nodes under the newly detected gateway.
                if bd.is_fam14():
                    await self._find_and_update_devices_belonging_to_gateway(bd.base_id)


    async def _find_and_update_devices_belonging_to_gateway(self, base_id:str):
        """Check all devices which are not detected as bus device (decentral/wireless device) if it belong to a gateway."""
        base_id_int = int.from_bytes( AddressExpression.parse(base_id)[0], 'big' )
        for d in self.devices.values():
            if not d.bus_device and d.base_id == '00-00-00-00':
                adr_int = int.from_bytes( AddressExpression.parse(d.address)[0], 'big' )
                if 0 < (adr_int - base_id_int) and (adr_int - base_id_int) < 0xFF:
                    d.address = data_helper.a2s(adr_int - base_id_int)
                    d.bus_device = True
                    d.base_id = base_id

                    self.app_bus.fire_event(AppBusEventType.UPDATE_DEVICE_REPRESENTATION, d)


    def get_device_by_id(self, device_id:str):
        for d in self.devices.values():
            if d.external_id == device_id:
                return d
            
        return None
    
    def get_sensors_configured_in_a_device(self, device:Device) -> list[Device]:
        """returns all sensors configured in a device"""
        sensors = []

        # if bus device selected
        for m in device.memory_entries:
            #for device with many addresses check if channel fits
            if device.channel == m.channel:

                # if sensor has global address
                if m.sensor_id_str in self.devices:
                    sensors.append(self.devices[m.sensor_id_str])

                else:
                    m_ext_id = data_helper.add_addresses(m.sensor_id_str, device.base_id)
                    if m_ext_id in self.devices:
                        sensors.append(self.devices[m_ext_id])

        return sensors
    
    
    def get_devices_containing_sensor_in_config(self, sensor:Device) ->list[Device]:
        """returns all devices which contain the given sensor in its config"""
        devices = []

        # if sensor selected
        for d in self.devices.values():
            for m in d.memory_entries:
                if sensor.address == m.sensor_id_str:
                    if d.channel == m.channel:
                        # sensor with global id
                        if m.sensor_id_str == sensor.external_id:
                            devices.append(d)
                        # sensor with local bus id
                        elif d.base_id == sensor.base_id:
                            devices.append(d)
        return devices


    def get_related_devices(self, device_external_id:str) -> list[Device]:
        """returns a list of all devices in which a sensor is entered or sensors which are configured inside an device."""
        if device_external_id is None or device_external_id == '' or device_external_id == 'Distributed Devices':
            return []

        device:Device = self.devices[device_external_id]
        
        devices = []
        devices.extend( self.get_sensors_configured_in_a_device(device) )
        devices.extend( self.get_devices_containing_sensor_in_config(device) )
        return devices

  

    def update_device(self, device: Device) -> None:
        self.devices[device.external_id] = device

        if device.is_bus_device():
            self.app_bus.fire_event(AppBusEventType.UPDATE_DEVICE_REPRESENTATION, device)
        else:
            self.app_bus.fire_event(AppBusEventType.UPDATE_SENSOR_REPRESENTATION, device)


    def find_device_by_local_address(self, address:str, base_id:str) -> Device:
        local_adr = int.from_bytes(AddressExpression.parse(address)[0], "big")
        base_adr = int.from_bytes(AddressExpression.parse(base_id)[0], "big")

        if (local_adr + base_adr) > 0xFFFFFFFF:
            return None

        ext_id = data_helper.a2s(local_adr + base_adr)
        if ext_id in self.devices:
            return self.devices[ext_id]
        
        return None


    def get_values_from_message_to_string(self, message:EltakoMessage, base_id:str=None) -> str:
        # get ext id
        ext_id_str = b2s(message.address)
        if ext_id_str.startswith('00-00-00-') and base_id is not None:
            device:Device = self.find_device_by_local_address(ext_id_str, base_id)
            if device is None: 
                return None, None
            else:
                ext_id_str = device.external_id

        eep = None
        if ext_id_str in self.devices:
            device = self.devices[ext_id_str]
            try:
                eep:EEP = data_helper.find_eep_by_name(device.eep)
                return eep, ', '.join(data_helper.get_values_for_eep(eep, message))
            except:
                pass
        return eep, None
    

    def on_update_send_message_template_list(self, data:list[str]):
        self.send_message_template_list = data
    


from tkinter import *
from tkinter import ttk
import webbrowser
from idlelib.tooltip import Hovertip

from ..icons.image_gallary import ImageGallery

class DonationButton(ttk.Button):

    def __init__(self, master=None, cnf={}, **kw):
        if 'small_icon' in kw and kw['small_icon']:
            image = ImageGallery.get_paypal_icon()
        else:
            image = ImageGallery.get_paypal_me_icon()
        
        super().__init__(master, image=image, cursor="hand2", command=self.open_payapl_me) # , relief=kw.get('relief', FLAT))
        self.image = image

        Hovertip(self,"Donate to support this project!",300)

        self.bind('<Return>', lambda e: self.open_payapl_me())

    def open_payapl_me(self):
        webbrowser.open_new(r"https://paypal.me/grimmpp")


import os
import sys
import unittest

file_dir = os.path.join( os.path.dirname(__file__), '..', 'eo_man', 'data')
sys.path.append(file_dir)
__import__('homeassistant') 

from eo_man.data.data_helper import EEP_MAPPING
from eo_man.data.const import CONF_EEP, CONF_TYPE

class TestGenerateDocs(unittest.TestCase):

    def test_generate_list_of_supported_devices(self):
        text = "# List of Supported Devices \n"
        text += "Devices not contained in this list are quite likely already supported by existing devices in this list. \n"

        text += "<br />\n"
        text += "(This list is auto-generated by using template definitions in [data_helper.py](https://github.com/grimmpp/enocean-device-manager/blob/main/eo_man/data/data_helper.py).) \n"
        text += "<br />\n"

        text += "| Brand | Name | Description | EEP | Sender EEP | HA Platform | \n"
        keys = ['brand', 'hw-type', 'description', CONF_EEP, 'sender_eep', CONF_TYPE]
        for k in keys:
            text += "|-----"
        text += "| \n"

        for e in EEP_MAPPING:
            if e.get('hw-type', '') is not 'BusObject':
                for k in keys:
                    text += f"| {e.get(k, '')} "
                text += "| \n"

        file='./docs/supported-devices.md' 
        with open(file, 'w') as filetowrite:
            filetowrite.write(text)


import tkinter as tk
from tkinter import ttk
from tkinter import *
from tkinter.font import *
import webbrowser
from tkinterhtml import HtmlFrame
from PIL import Image, ImageTk
from idlelib.tooltip import Hovertip
from tkscrolledframe import ScrolledFrame

from ..data.const import *

from .donation_button import DonationButton

from ..data.app_info import ApplicationInfo as AppInfo
from ..data.data_helper import EEP_MAPPING


class DeviceInfoWindow():

    def __init__(self, main:Tk):
        popup = Toplevel(main)
        popup.wm_title("Supported Devices")
        self.popup = popup

        self.label_font = Font(name="Arial", underline=True, size=10)

        l = tk.Label(popup, text="Device Information", font='Arial 14 bold')
        l.pack(side=TOP, fill="x", pady=10)

        scrolledFrame = ScrolledFrame(popup, use_ttk=True)
        scrolledFrame.pack(side=LEFT, fill=BOTH, expand=2)

        # Bind the arrow keys and scroll wheel
        scrolledFrame.bind_arrow_keys(popup)
        scrolledFrame.bind_scroll_wheel(popup)

        table = scrolledFrame.display_widget(ttk.Treeview)
        table.pack(expand=True, fill=BOTH)

        table['columns'] = ('#1','#2', '#3', '#4', '#5', '#6', '#7')
        table["show"] = "headings"
        table.heading('#1', text='Brand')
        table.heading('#2', text='Device')
        table.heading('#3', text='Description')
        table.heading('#4', text='EEP (Sensor)')
        table.heading('#5', text='Sender EEP (Controller)')
        table.heading('#6', text='HA Platform')
        table.heading('#7', text='Documentation')

        for m in EEP_MAPPING:
            table.insert('', tk.END, values=(
                m.get('brand', ''), 
                m.get('hw-type', ''),
                m.get('description', ''),
                m.get(CONF_EEP, ''),
                m.get('sender_eep', ''),
                m.get(CONF_TYPE, ''),
                m.get('docs', '')
            ))
            


        popup.wm_attributes('-toolwindow', 'True')
        popup.resizable (width=True, height=True)
        popup.transient(main)

        popup.state('zoomed')

        main.wait_window(popup)



from datetime import datetime

from homeassistant.const import CONF_ID, CONF_DEVICES, CONF_NAME

from ..controller.app_bus import AppBus, AppBusEventType

from .app_info import ApplicationInfo as AppInfo
from .data_manager import DataManager
from .device import Device
from .const import *
from . import data_helper

class HomeAssistantConfigurationGenerator():

    LOCAL_SENDER_OFFSET_ID = '00-00-B0-00'

    def __init__(self, app_bus:AppBus, data_manager:DataManager):
        self.app_bus = app_bus
        self.data_manager = data_manager

    def get_description(self) -> str:
        return f"""
# DESCRIPTION:
#
# This is an automatically generated Home Assistant Configuration for the Eltako Integration (https://github.com/grimmpp/home-assistant-eltako)
# Generated at {datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f")} and by: 'EnOcean Device Manager':
{AppInfo.get_app_info_as_str(prefix='# ')}#
# Hints:
# * The auto-generation considers all devices which are marked with 'Export to HA' = True. 
# * Decentralized devices are entered into the list under devices for every gateway. If you are using more than one gateway you probably want to have those only once listed. Please remove "dupplicated" entries.
# * FAM14 gatways can be easily exchanged by FGW14-USB gateways. You just need to change the value of 'device_type' from 'fam14' to 'fgw14usb'.
# * 'id' of the gateways are random and are just counted up. You can simply change them if needed. Just ensure that they are unique and not specified more than once. 
#
"""

    def get_gateway_by(self, gw_d:Device) -> GatewayDeviceType:
        gw_type = None
        for t in GatewayDeviceType:
            if GATEWAY_DISPLAY_NAMES[t.value].lower() in gw_d.device_type.lower():
                return t.value
        return None

    def perform_tests(self) -> str:
        device_list = [d for d in self.data_manager.devices.values() if not d.is_gateway() and d.use_in_ha]

        try:
            self.test_unique_sender_ids(device_list)
            return None
        except Exception as e:
            return str(e)

    def test_unique_sender_ids(self, device_list:list[Device]):
        sender_ids = {}
        for d in device_list:
            if 'sender' in d.additional_fields and 'id' in d.additional_fields['sender']:
                sender_id = d.additional_fields['sender']['id']
                if int(sender_id, 16) < 1 or int(sender_id, 16) > 127:
                    raise Exception(f"sender id '{sender_id}' of device '{d.external_id}' is no valid number between 1 and 127.")
                if sender_id in sender_ids:
                    raise Exception(f"sender id '{sender_id}' is assigned more than once for at least device '{sender_ids[sender_id].external_id}' and '{d.external_id}'.")
                
                sender_ids[sender_id] = d

    def generate_ha_config(self, device_list:list[Device]) -> str:
        ha_platforms = set([str(d.ha_platform) for d in device_list if d.ha_platform is not None])
        gateways = [d for d in device_list if d.is_gateway() and d.use_in_ha]
        devices = [d for d in device_list if not d.is_gateway() and d.use_in_ha]

        out = self.get_description()
        out += "\n"
        out += f"{DOMAIN}:\n"
        out += f"  {CONF_GERNERAL_SETTINGS}:\n"
        out += f"    {CONF_FAST_STATUS_CHANGE}: False\n"
        out += f"    {CONF_SHOW_DEV_ID_IN_DEV_NAME}: False\n"
        out += f"\n"
        
        global_gw_id = 0
        # add fam14 gateways
        for gw_d in gateways:
            global_gw_id += 1
            out += f"  {CONF_GATEWAY}:\n"
            out += f"  - {CONF_ID}: {global_gw_id}\n"

            gw_fam14 = GatewayDeviceType.EltakoFAM14.value
            gw_fgw14usb = GatewayDeviceType.EltakoFGW14USB.value
            
            # gw_type = self.get_gateway_by(gw_d)
            out += f"    {CONF_DEVICE_TYPE}: {gw_d.device_type}   # you can simply change {gw_fam14} to {gw_fgw14usb}\n"
            out += f"    {CONF_BASE_ID}: {gw_d.base_id}\n"
            out += f"    # {CONF_COMMENT}: {gw_d.comment}\n"
            if gw_d.device_type == GatewayDeviceType.LAN:
                out += f"    {CONF_GATEWAY_ADDRESS}: {gw_d.additional_fields['address']}\n"
            out += f"    {CONF_DEVICES}:\n"

            for platform in ha_platforms:
                if platform != '':
                    out += f"      {platform}:\n"
                    for device in devices:
                        if device.ha_platform == platform:
                            # add devices
                            out += self.config_section_from_device_to_string(gw_d, device, True, 0) + "\n\n"
        # logs
        out += "\n"
        out += "logger:\n"
        out += "  default: info\n"
        out += "  logs:\n"
        out += f"    {DOMAIN}: info\n"

        return out



    def config_section_from_device_to_string(self, gateway:Device, device:Device, is_list:bool, space_count:int=0) -> str:
        out = ""
        spaces = space_count*" " + "        "

        # add user comment
        if device.comment:
            out += spaces + f"# {device.comment}\n"
        
        # list related devices for comment
        rel_devs = []
        for d in self.data_manager.get_related_devices(device.external_id):
            rel_devs.append( f"{d.name} (Type: {d.device_type}, Adr: {d.address})" )
        if len(rel_devs):
            out += spaces + f"# Related devices: {', '.join(rel_devs)}\n"
            
        info = data_helper.find_device_info_by_device_type(device.device_type)
        if info and 'PCT14-key-function' in info:
            kf = info['PCT14-key-function']
            fg = info['PCT14-function-group']
            out += spaces[:-2] + f"  # Use 'Write HA senders to devices' button or enter manually sender id in PCT14 into function group {fg} with function {kf} \n"    
        adr = device.address if gateway.is_wired_gateway() and gateway.base_id == device.base_id else device.external_id
        out += spaces[:-2] + f"- {CONF_ID}: {adr.strip()}\n"
        out += spaces[:-2] + f"  {CONF_NAME}: {device.name}\n"
        out += spaces[:-2] + f"  {CONF_EEP}: {device.eep}\n"

        out += self.export_additional_fields(gateway, device.additional_fields, space_count)
        
        return out
    

    def export_additional_fields(self, gateway:Device, additional_fields:dict, space_count:int=0, parent_key:str=None) -> str:
        out = ""
        spaces = space_count*" " + "        "
        for key in additional_fields.keys():
            value = additional_fields[key]
            if parent_key in ['sender'] and key == 'id':
                if gateway.is_wired_gateway(): sender_offset_id = self.LOCAL_SENDER_OFFSET_ID
                else: sender_offset_id = gateway.base_id
                value = data_helper.a2s( int("0x"+value[-2:], base=16) + data_helper.a2i(sender_offset_id) )
            if isinstance(value, str) or isinstance(value, int):
                if key not in [CONF_COMMENT, CONF_REGISTERED_IN]:
                    if isinstance(value, str) and '?' in value:
                        value += " # <= NEED TO BE COMPLETED!!!"
                    out += f"{spaces}{key}: {value}\n"
            elif isinstance(value, dict):
                out += f"{spaces}{key}: \n"
                out += self.export_additional_fields(gateway, value, space_count+2, key)
        return out
    

    def save_as_yaml_to_file(self, filename:str):
        
        devices = self.data_manager.devices.values()

        config_str = self.generate_ha_config(devices)

        with open(filename, 'w', encoding="utf-8") as f:
            print(config_str, file=f)



import tkinter as tk
from tkinter import ttk
from tkinter import *
from tkinter.font import *
import webbrowser
from tkinterhtml import HtmlFrame
from idlelib.tooltip import Hovertip

from .donation_button import DonationButton

from ..data.app_info import ApplicationInfo as AppInfo


class AboutWindow():

    def __init__(self, main:Tk):
        popup = Toplevel(main)
        popup.wm_title("About")
        self.popup = popup

        self.label_font = Font(name="Arial", underline=True, size=10)

        l = tk.Label(popup, text="About EnOcean Device Manager", font='Arial 14 bold')
        l.pack(side=TOP, fill="x", pady=10)

        l = tk.Label(popup, text="Version: "+AppInfo.get_version(), font="Arial 10", anchor="w")
        l.pack(side=TOP, fill="x", pady=2, padx=5)

        self.get_label_link(text="PyPI: Package Name: "+__package__.split('.')[0], link=r"https://pypi.org/project/eo-man")
        
        self.get_label_link(text="GitHub: grimmpp/enocean-device-manager", link=AppInfo.get_home_page())

        self.get_label_link(text="GitHub: EnOcean Device Manager Documentation", link=r"https://github.com/grimmpp/enocean-device-manager/tree/main/docs")

        self.get_label_link(text="GitHub: Report a bug or ask for features!", link=AppInfo.get_home_page()+"/issues")

        self.get_label_link(text="Author: "+AppInfo.get_author(), link=r"https://github.com/grimmpp")

        self.get_label_link(text="License - "+AppInfo.get_license(), link=r"https://github.com/grimmpp/enocean-device-manager/blob/main/LICENSE")

        b = DonationButton(popup)
        b.pack(side=TOP, pady=(8,0), padx=5)
        b.focus()

        b = ttk.Button(popup, text="OK", command=popup.destroy)
        b.bind('<Return>', lambda e: popup.destroy())
        b.pack(side=TOP, fill="x", pady=(10,2), padx=10 )
        # b.focus()

        popup.wm_attributes('-toolwindow', 'True')
        popup.resizable (width=False, height=False)
        popup.transient(main)
        popup.grab_set()

        # center
        w = 340
        h = 310
        x = main.winfo_x() + main.winfo_width()/2 - w/2
        y = main.winfo_y() + main.winfo_height()/2 - h/2
        popup.geometry('%dx%d+%d+%d' % (w, h, x, y))

        main.wait_window(popup)

    def callback(self, url):
        webbrowser.open(url)

    def get_label_link(self, text:str, link:str, tooltip:str=None) -> Label:
        if tooltip is None:
            tooltip=link

        l = tk.Label(self.popup, text=text, fg="blue", font=self.label_font, cursor="hand2", anchor="w")
        l.pack(side=TOP, fill="x", pady=2, padx=5)
        l.bind("<Button-1>", lambda e: self.callback(link))
        Hovertip(l,tooltip,300)



import os
import sys
import logging

LOGGER = logging.getLogger()

def load_dep_homeassistant():
    # import fake homeassistant package
    file_dir = os.path.join( os.path.dirname(__file__), 'data')
    sys.path.append(file_dir)
    __import__('homeassistant')


from datetime import datetime
from eltakobus.message import ESP2Message

class RecordedMessage():

    def __init__(self, message:ESP2Message, external_id:str, gateway_id:str) -> None:
        self.message:ESP2Message = None
        self.external_device_id:str = external_id
        self.received_via_gateway_id: str = gateway_id
        self.received = datetime.now()


import logging

from tkinter import *
from tkinter import ttk

from ..icons.image_gallary import ImageGallery

from ..controller.app_bus import AppBus, AppBusEventType
from ..controller.serial_controller import SerialController

from ..data.data_manager import DataManager

from ..view import DEFAULT_WINDOW_TITLE
from ..view.device_details import DeviceDetails
from ..view.device_table import DeviceTable
from ..view.filter_bar import FilterBar
from ..view.log_output import LogOutputPanel
from ..view.menu_presenter import MenuPresenter
from ..view.serial_communication_bar import SerialConnectionBar
from ..view.status_bar import StatusBar
from ..view.tool_bar import ToolBar

class MainPanel():

    def __init__(self, app_bus:AppBus, data_manager: DataManager):
        self.main = Tk()
        self.app_bus = app_bus
        ## init main window
        self._init_window()

        ## define grid
        row_button_bar = 0
        row_serial_con_bar = 1
        row_filter_bar = 2
        row_main_area = 3
        row_status_bar = 4
        self.main.rowconfigure(row_button_bar, weight=0, minsize=38)      # button bar
        self.main.rowconfigure(row_serial_con_bar, weight=0, minsize=38)      # serial connection bar
        self.main.rowconfigure(row_filter_bar, weight=0, minsize=38)      # table filter bar
        self.main.rowconfigure(row_main_area, weight=5, minsize=100)     # treeview
        # main.rowconfigure(2, weight=1, minsize=30)    # logview
        self.main.rowconfigure(row_status_bar, weight=0, minsize=30)      # status bar
        self.main.columnconfigure(0, weight=1, minsize=100)

        serial_controller = SerialController(app_bus)

        ## init presenters
        mp = MenuPresenter(self.main, app_bus, data_manager, serial_controller)
        
        ToolBar(self.main, mp, row=row_button_bar)
        SerialConnectionBar(self.main, app_bus, data_manager, serial_controller, row=row_serial_con_bar)
        FilterBar(self.main, app_bus, data_manager, row=row_filter_bar)
        # main area
        main_split_area = ttk.PanedWindow(self.main, orient=VERTICAL)
        main_split_area.grid(row=row_main_area, column=0, sticky=NSEW, columnspan=4)
        
        data_split_area = ttk.PanedWindow(main_split_area, orient=HORIZONTAL)
        # data_split_area = Frame(main_split_area)
        # data_split_area.columnconfigure(0, weight=5)
        # data_split_area.columnconfigure(0, weight=0, minsize=100)
        
        dt = DeviceTable(data_split_area, app_bus, data_manager)
        dd = DeviceDetails(self.main, data_split_area, app_bus, data_manager)
        lo = LogOutputPanel(main_split_area, app_bus, data_manager)

        main_split_area.add(data_split_area, weight=5)
        main_split_area.add(lo.root, weight=2)

        data_split_area.add(dt.root, weight=5)
        data_split_area.add(dd.root, weight=0)
        # dt.root.grid(row=0, column=0, sticky="nsew")
        # dd.root.grid(row=0, column=1, sticky="nsew")

        StatusBar(self.main, app_bus, data_manager, row=row_status_bar)

        self.main.after(1, lambda: self.main.focus_force())

        ## start main loop
        self.main.mainloop()

        
        


    def _init_window(self):
        self.main.title(DEFAULT_WINDOW_TITLE)

        #style
        style = ttk.Style()
        style.configure("TButton", relief="sunken", background='green')
        style_theme = 'xpnative' # 'clam'
        self.app_bus.fire_event(AppBusEventType.LOG_MESSAGE, {'msg': f"Available style themes: {ttk.Style().theme_names()}", 'log-level': 'DEBUG'})
        try:
            style.theme_use(style_theme)
        except:
            self.app_bus.fire_event(AppBusEventType.LOG_MESSAGE, {'msg': f"Cannot load style theme {style_theme}!", 'log-level': 'WARNING'})

        self.main.geometry("1400x600")  # set starting size of window
        # self.main.attributes('-fullscreen', True)
        # self.main.state('zoomed') # opens window maximized

        self.main.config(bg="lightgrey")
        self.main.protocol("WM_DELETE_WINDOW", self.on_closing)

        # icon next to title in window frame
        self.main.wm_iconphoto(False, ImageGallery.get_eo_man_logo())

        # icon in taskbar
        icon = ImageGallery.get_eo_man_logo()
        self.main.iconphoto(True, icon, icon)

    def on_loaded(self) -> None:
        self.app_bus.fire_event(AppBusEventType.WINDOW_LOADED, {})

    def on_closing(self) -> None:
        self.app_bus.fire_event(AppBusEventType.WINDOW_CLOSED, {})
        logging.info("Close Application eo-man")
        logging.info("========================\n")
        self.main.destroy()


from typing import Final
import yaml

from .device import Device
from .filter import DataFilter
from .recorded_message import RecordedMessage

import pickle

class ApplicationData():

    class_version:Final = '1.0.0'

    def __init__(self, version:str='unknown', 
                 selected_data_filter:str=None, data_filters:dict[str:DataFilter]={},
                 devices:dict[str:Device]={},
                 recoreded_messages:list[RecordedMessage]=[]):
        
        self.application_version:str = version

        self.selected_data_filter_name:str=selected_data_filter
        self.data_filters:dict[str:DataFilter] = data_filters

        self.devices:dict[str:Device] = devices

        self.recoreded_messages:list[RecordedMessage] = recoreded_messages

        self.send_message_template_list: list[str] = []


    translations:dict[str:str] ={
        'name: HA Contoller': 'name: HA Controller',
        '(Wireless Tranceiver)': '(Wireless Transceiver)'
    }

    @classmethod
    def read_from_file(cls, filename:str):
        result = ApplicationData()

        file_content = None
        with open(filename, 'rb') as file:
            file_content = pickle.loads(file) 

        if isinstance(file_content, ApplicationData):
            result = file_content
            return result
        
        # to be downwards compatible
        if isinstance(file_content, dict) and len(file_content) > 0 and isinstance(list(file_content.values())[0], Device):
            result.devices = file_content

        if hasattr(file_content, 'devices'):
            result.devices = file_content.devices

        if hasattr(file_content, 'data_filters'):
            result.data_filters = file_content.data_filters
            
        if hasattr(file_content, 'selected_data_filter_name'):
            result.selected_data_filter_name = file_content.selected_data_filter_name

        if hasattr(file_content, 'application_version'):
            result.application_version = file_content.application_version

        return result
    
    @classmethod
    def _migrate(cls, obj):
        """required to make different versions compatibel"""
        if not hasattr(obj, 'recoreded_messages'):
            setattr(obj, 'recoreded_messages', [])

        if not hasattr(obj, 'send_message_template_list'):
            setattr(obj, 'send_message_template_list', [])

    @classmethod
    def read_from_yaml_file(cls, filename:str):
        with open(filename, 'r') as file:
            file_content = file.read()
            for k,v in cls.translations.items():
                file_content.replace(k,v)
            app_data = yaml.load(file_content, Loader=yaml.Loader)
        cls._migrate(app_data)
        
        return app_data
        
    
    # @classmethod
    # def from_yaml(cls, constructor, node):
    #     return cls(version=node.version, 
    #                selected_data_filter=node.selected_data_filter,
    #                data_filters=node.data_filters,
    #                devices=node.devices
    #                )

    @classmethod
    def write_to_file(cls, filename:str, application_data):
        with open(filename, 'wb') as file:
            pickle.dump(application_data, file)

    @classmethod
    def write_to_yaml_file(cls, filename:str, application_data):
        with open(filename, 'w') as file:
            yaml.dump(application_data, file)